#include <err.h>
#include <fenv.h>
#include <gmp.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "wrapper.h"

size_t coeff_ml_len(unsigned int flags, const ap_coeff_t *c) {
	size_t l;
	mpq_t m;

	mpq_init(m);
	coeff2mpq(m, c);

	l = mpq_len(m, flags);
	mpq_clear(m);
	return l;
}

char *coeff_ml_pretty(char *s, unsigned int flags, const ap_coeff_t *c) {
	mpq_t m;

	mpq_init(m);
	coeff2mpq(m, c);

	s = mpq_pretty(s, m, flags);
	mpq_clear(m);
	return s;
}

void coeff2mpq(mpq_t m, const ap_coeff_t *c) {
	const ap_scalar_t *s;

	switch(c->discr) {
	case AP_COEFF_SCALAR:
		s =  c->val.scalar;
		break;

	case AP_COEFF_INTERVAL:
		errx(1, "coeff is an interval");

	default:
		errx(1, "coeff is of unknown type");
	}

	switch(s->discr) {
	case AP_SCALAR_MPQ:
		mpq_set(m, s->val.mpq);
		break;

	case AP_SCALAR_DOUBLE:
		mpq_set_d(m, s->val.dbl);
		break;

	case AP_SCALAR_MPFR:
		errx(1, "scalar is an mpfr number");

	default:
		errx(1, "scalar is of unknown type");
	}
}
