#include <err.h>
#include <gmp.h>
#include <string.h>

#include "wrapper.h"

static const char *cons_beg(ap_constyp_t);

size_t
lcons_len(const ap_lincons1_t *l, struct env *e)
{
	const char *begin;
	const ap_lincons0_t *cons;
	size_t len;

	env_merge(e, l->env);

	cons = &(l->lincons0);

	begin = cons_beg(cons->constyp);
	len = strlen(begin);
	len += lexpr_len(cons->linexpr0, e);
	len += 2; /* space */
	len += coeff_ml_len(C_PAREN, &(cons->linexpr0->cst));

	return len;
}

char *
lcons_pretty(char *s, const ap_lincons1_t *l, struct env *e)
{
	const ap_lincons0_t *cons;

	cons = &(l->lincons0);

	s = frag_pretty(s, cons_beg(cons->constyp));
	s = frag_pretty(s, " ");
	s = lexpr_pretty(s, cons->linexpr0, e);
	s = frag_pretty(s, " ");
	s = coeff_ml_pretty(s, C_PAREN, &(cons->linexpr0->cst));

	return s;
}

unsigned int
tconsArEqs(ap_tcons0_array_t *a)
{
	unsigned int i, eqs = 0;

	for(i = 0; i < a->size; i++)
		if(cons_typ(a->p[i].constyp) == L_EQ)
			eqs++;

	return eqs;
}

unsigned int
tconsArIneqs(ap_tcons0_array_t *a)
{
	unsigned int i, ineqs = 0;

	for(i = 0; i < a->size; i++)
		if(cons_typ(a->p[i].constyp) == L_INEQ)
			ineqs++;

	return ineqs;
}

size_t
tcons_len(ap_environment_t *env, ap_tcons0_t *t, struct env *e)
{
	const char *beg;
	size_t sz = 0;
	mpq_t cst;

	beg = cons_beg(t->constyp);
	sz += strlen(beg);
	sz += texpr_len(C_NEG, env, t->texpr0, e);
	sz += 2; /* spaces */

	mpq_init(cst);
	texpr_cst(cst, t->texpr0);
	sz += mpq_len(cst, C_PAREN);

	mpq_clear(cst);
	return sz;
}

char *
tcons_pretty(char *s, ap_environment_t *env, ap_tcons0_t *t, struct env *e)
{
	const char *beg;
	mpq_t cst;

	beg = cons_beg(t->constyp);
	s = frag_pretty(s, beg);
	s = frag_pretty(s, " ");
	s = texpr_pretty(s, C_NEG, env, t->texpr0, e);
	s = frag_pretty(s, " ");

	mpq_init(cst);
	texpr_cst(cst, t->texpr0);
	s = mpq_pretty(s, cst, C_PAREN);

	mpq_clear(cst);
	return s;
}

static const char *cons_beg(ap_constyp_t t) {
	switch(t) {
	case AP_CONS_EQ:
		return "\t\tCstr.mkInt Cmp.Eq";

	case AP_CONS_SUPEQ:
		return "\t\tCstr.mkInt Cmp.Le";

	case AP_CONS_SUP:
		return "\t\tCstr.mkInt Cmp.Lt";

	case AP_CONS_EQMOD:
		errx(1, "AP_CONS_EQMOD unsupported");

	case AP_CONS_DISEQ:
		errx(1, "AP_CONS_DISEQ unsupported");

	default:
		errx(1, "unknown ap_constyp_t");
	}
}

unsigned int cons_typ(ap_constyp_t t) {
	switch(t) {
	case AP_CONS_EQ:
		return L_EQ;

	case AP_CONS_SUPEQ:
		return L_INEQ;

	case AP_CONS_SUP:
		return L_INEQ;

	case AP_CONS_EQMOD:
		errx(1, "AP_CONS_EQMOD unsupported");

	case AP_CONS_DISEQ:
		errx(1, "AP_CONS_DISEQ unsupported");

	default:
		errx(1, "unknown ap_constyp_t");
	}
}
