#include <err.h>
#include <stdio.h>
#include <stdlib.h>

#include "wrapper.h"

const char *inclcall = "incl%u ();\n";

static const char *inclopb =
	"(* Db.Apron ({ Db.prg = \"#\"; Db.op = Db.Incl; Db.n = %u }, Int64.of_float (%lf)); *)\n"
	"(* Db.Eqs ({ Db.prg = \"#\"; Db.op = Db.Incl; Db.n = %u }, [%u; %u]); *)\n"
	"(* Db.Ineqs ({ Db.prg = \"#\"; Db.op = Db.Incl; Db.n = %u }, [%u; %u]); *)\n"
	"let incl%u () =\n"
	"\tlet reftime = Int64.of_float (%lf) in\n"
	"\tlet refans = %s in\n";

static const char *inclope =
	"\tHelper.incl_out %u reftime refans p%u p%u\n;;\n\n";

void inclrec(struct record *r, ap_manager_t *m, double reftime,
		bool refans, ap_abstract1_t *a1, ap_abstract1_t *a2) {
	size_t sz;
	char *s, *e;
	struct env env;
	ap_lincons1_array_t ar1, ar2;

	env_new(&env);
	ar1 = ap_abstract1_to_lincons_array(m, a1);
	ar2 = ap_abstract1_to_lincons_array(m, a2);
	env_merge(&env, ar1.env);
	env_merge(&env, ar2.env);

	sz = pollenML(L_SZ, 0, &ar1, &env);
	sz += pollenML(L_SZ, 1, &ar2, &env);
	sz += env_len(L_SZ, &env);

	fprintf(r->f, inclopb, r->cnt, reftime, r->cnt, polEqs(&ar1), polEqs(&ar2),
		r->cnt, polIneqs(&ar1), polIneqs(&ar2), r->cnt, reftime, refans ? "true" : "false");

	if((s = (char *)malloc(sz + 2)) == NULL)
		err(1, "malloc(3)");

	e = s;
	e = env_pretty(e, L_SZ, &env);
	e = polprettyML(e, L_SZ, 0, &ar1, &env);
	e = polprettyML(e, L_SZ, 1, &ar2, &env);
	*e = '\0';

	if(sz < e - s)
		warnx("inclfrec(): %i, %i", e - s, sz);

	fprintf(r->f, "%s\n", s);
	free((void *)s);

	fprintf(r->f, inclope, r->cnt, 0, 1);

	r->cnt++;
	ap_lincons1_array_clear(&ar1);
	ap_lincons1_array_clear(&ar2);
	env_del(&env);
}
