#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "wrapper.h"

const char *meet_call = "meet%u ();\n";

static const char *meet_beg =
	"(* Db.Apron ({ Db.prg = \"#\"; Db.op = Db.Meet; Db.n = %u }, Int64.of_float (%lf)); *)\n"
	"(* Db.Eqs ({ Db.prg = \"#\"; Db.op = Db.Meet; Db.n = %u }, [%u; %u]); *)\n"
	"(* Db.Ineqs ({ Db.prg = \"#\"; Db.op = Db.Meet; Db.n = %u }, [%u; %u]); *)\n"
	"let meet%u () =\n"
	"\tlet reftime = Int64.of_float (%lf) in\n";

static const char *cons_beg = "\tlet l = [\n";
static const char *cons_end = "\n\t] in\n";

void
meetrec(struct record *r, ap_manager_t *m, double reftime, ap_abstract1_t *a,
	ap_tcons1_array_t *t, ap_abstract1_t *res)
{
	char *s1, *s2;
	unsigned int u;
	size_t sz = 0;
	ap_lincons1_array_t ar, arres;
	ap_tcons0_array_t t0;
	ap_dimchange_t *dc = NULL;

	struct env e;
	if(t->tcons0_array.size == 0) {
		fprintf(r->f, "let meet%u () = ()\n\n", r->cnt++);
		return;
	}

	if(!ap_environment_is_eq(a->env, t->env)) {
		if((dc = ap_environment_dimchange(t->env, a->env)) == NULL)
			errx(1, "meetrec(): ap_environment_dimchange()");

		t0 = ap_tcons0_array_add_dimensions(&t->tcons0_array, dc);
	} else
		t0 = t->tcons0_array;

	env_new(&e);
	env_merge(&e, a->env);
	env_merge(&e, res->env);

	ar = ap_abstract1_to_lincons_array(m, a);
	sz += pollenML(L_SZ, 0, &ar, &e);

	arres = ap_abstract1_to_lincons_array(m, res);
	sz += pollenML(L_STAT, 1, &arres, &e);

	sz += strlen(cons_beg);
	for(u = 0; u < t0.size; u++) {
		sz += tcons_len(t->env, &t0.p[u], &e);
		sz += 2; /* ";\n" */
	}
	sz += strlen(cons_end);

	sz += env_len(L_SZ, &e);

	if((s1 = (char *)malloc(sz + 2)) == NULL)
		err(1, "malloc(3)");

	s2 = env_pretty(s1, L_SZ, &e);
	s2 = polprettyML(s2, L_STAT, 1, &arres, &e);
	s2 = polprettyML(s2, L_SZ, 0, &ar, &e);
	s2 = frag_pretty(s2, cons_beg);
	for(u = 0; u < t0.size; u++) {
		if(u)
			s2 = frag_pretty(s2, ";\n");

		s2 = tcons_pretty(s2, t->env, &t0.p[u], &e);
	}
	s2 = frag_pretty(s2, cons_end);

	if(sz < s2 - s1)
		warnx("meetrec(): estimated: %i, actual: %i", sz, s2 - s1);

	fprintf(r->f, meet_beg, r->cnt, reftime, r->cnt, polEqs(&ar), tconsArEqs(&t0),
		r->cnt, polIneqs(&ar), tconsArIneqs(&t0), r->cnt, reftime);
	fprintf(r->f, "%s\tHelper.meet_out %u reftime p1 p1Eqs p1Ineqs p0 l\n\n",
		s1, r->cnt);

	r->cnt++;

	if(dc) {
		ap_dimchange_free(dc);
		ap_tcons0_array_clear(&t0);
	}

	ap_lincons1_array_clear(&ar);
	ap_lincons1_array_clear(&arres);
	free((void *)s1);
	env_del(&e);
}
