#include <err.h>
#include <fenv.h>
#include <gmp.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "wrapper.h"

static const int gmp_base = 10;
static const char *ml_coeff = "Num.mk ~den:(%lu) (%li)";
static const char *ml_coeff2 = "Num.mk (%li)";
static const char *ml_coeffp = "(Num.mk ~den:(%lu) (%li))";
static const char *ml_coeffp2 = "(Num.mk (%li))";
static const char *ml_coeffz = "Num.z";
static const char *ml_coeffu = "Num.u";

static size_t mpz_len(mpz_t);

size_t
mpq_len(mpq_t m, unsigned int flags)
{
	long int n;
	size_t l;

	if((n = mpz_get_si(mpq_numref(m))) == 0)
		return flags & C_SKIP ? 0 : strlen(ml_coeffz);

	n = flags & C_NEG ? -n : n;
	if(mpz_get_ui(mpq_denref(m)) == 1) {
		if(n == 1)
			return strlen(ml_coeffu);

		l = strlen(flags & C_PAREN ? ml_coeffp2 : ml_coeff2) - 3;
		l += mpz_len(mpq_numref(m));
		return l;
	}

	l = strlen(flags & C_PAREN ? ml_coeffp : ml_coeff) - 6;
	l += mpz_len(mpq_numref(m));
	l += mpz_len(mpq_denref(m));

	return l;
}

char *
mpq_pretty(char *s, mpq_t m, unsigned int flags)
{
	long int n;
	long unsigned int d;

	if((n = mpz_get_si(mpq_numref(m))) == 0) {
		if(flags & C_SKIP)
			return 0;
		else
			return frag_pretty(s, ml_coeffz);
	}

	n = flags & C_NEG ? -n : n;
	if((d = mpz_get_ui(mpq_denref(m))) == 1) {
		if(n == 1)
			s += sprintf(s, ml_coeffu);
		else
			s += sprintf(s, flags & C_PAREN ? ml_coeffp2 : ml_coeff2, n);
	} else
		s += sprintf(s, flags & C_PAREN ? ml_coeffp : ml_coeff, d, n);

	return s;
}

/* GMP documentation says two is a safe minimum to add, but that is counting
the terminating 0 byte, which is accounted for higher in the call stack. */
static size_t mpz_len(mpz_t p) {
	if(mpz_fits_slong_p(p) == 0)
		errx(1, "too large to fit in a long");

	return mpz_sizeinbase(p, gmp_base) + 1;
}
