type 'n t =
| Nil
| Sub of ('n t * 'n * 'n t)

type 'n rtree_t = 'n t

let rec get (z: 'n) (rt: 'n rtree_t) (v: Var.t) =
	match rt with
	| Nil -> z
	| Sub (l, n, r) ->
		match v with
		| Var.XH -> n
		| Var.XO t -> get z l t
		| Var.XI t -> get z r t

let set (z: 'n) (rt0: 'n rtree_t) (v0: Var.t) (n0: 'n) =
	let rec _set rt v: 'n rtree_t =
		match rt, v with
		| Nil, Var.XH -> Sub (Nil, n0, Nil)
		| Sub (l, _, r), Var.XH -> Sub (l, n0, r)
		| Nil, Var.XO t -> Sub (_set Nil t, z, Nil)
		| Sub (l, n, r), Var.XO t -> Sub (_set l t, n, r)
		| Nil, Var.XI t -> Sub (Nil, z, _set Nil t)
		| Sub (l, n, r), Var.XI t -> Sub (l, n, _set r t)
	in
		_set rt0 v0

let mk z l = List.fold_left (fun a (x, n) -> set z a x n) Nil l

let map (f: 'n -> 'm) (rt0: 'n rtree_t) =
	let rec _map rt =
		match rt with
		| Nil -> Nil
		| Sub (l, n, r) -> Sub (_map l, f n, _map r)
	in
		_map rt0

let rec fold (f: 'a -> 'n -> 'a) (a: 'a) (rt: 'n rtree_t): 'a =
	match rt with
	| Nil -> a
	| Sub (l, n, r) -> fold f (fold f (f a n) l) r

let rec find (f: 'a -> 'b option): 'a t -> (Var.t * 'b) option = function
	| Nil -> None
	| Sub (l, n, r) ->
		match f n with
		| Some b -> Some (Var.XH, b)
		| None ->

		match find f l with
		| Some (v, b) -> Some (Var.XO v, b)
		| None ->

		match find f r with
		| Some (v, b) -> Some (Var.XI v, b)
		| None -> None

let pr sep npp ba0 rt0 =
	let dosep n l r =
		match n, l, r with
		| s, "", "" | "", s, "" | "", "", s -> s
		| s1, s2, "" | s1, "", s2 | "", s1, s2 -> s1 ^ sep ^ s2
		| s1, s2, s3 -> s1 ^ sep ^ s2 ^ sep ^ s3
	in
	let rec _pp ba rt =
		match ba, rt with
		| _, Nil -> ""
		| Nil, Sub (l, n, r) -> dosep (npp "?" n) (_pp Nil l) (_pp Nil r)
		| Sub (l1, n1, r1), Sub (l2, n2, r2) ->
			dosep (npp n1 n2) (_pp l1 l2) (_pp r1 r2)
	in
		_pp ba0 rt0
