Require Import QArith.
Require Import Qcanon.
Require CoqAddOn.

Definition t: Set := Qc.

Definition z: t := 0.
Definition u: t := 1.

Definition Le: t -> t -> Prop := Qcle.
Definition Lt: t -> t -> Prop := Qclt.

Definition add: t -> t -> t := Qcplus.
Definition sub: t -> t -> t := Qcminus.
Definition mul: t -> t -> t := Qcmult.
Definition div: t -> t -> t := Qcdiv.

Definition neg: t -> t := Qcopp.
Definition inv: t -> t := Qcinv.

Lemma EqDec: forall n1 n2: t, { n1 = n2 } + { n1 <> n2 }.
Proof Qc_eq_dec.

Lemma LtLeDec: forall n1 n2: t, { Lt n1 n2 } + { Le n2 n1 }.
Proof Qclt_le_dec.

Lemma LtLe: forall n1 n2: t, Lt n1 n2 -> Le n1 n2.
Proof Qclt_le_weak.

Lemma LtNotLe: forall n1 n2: t, Lt n1 n2 -> not (Le n2 n1).
Proof Qclt_not_le.

Lemma LeNotLt: forall n1 n2: t, Le n1 n2 -> not (Lt n2 n1).
Proof Qcle_not_lt.

Lemma LeTrans: forall n1 n2 n3: t, Le n1 n2 -> Le n2 n3 -> Le n1 n3.
Proof Qcle_trans.

Lemma LeLtTrans: forall n1 n2 n3: t, Le n1 n2 -> Lt n2 n3 -> Lt n1 n3.
Proof Qcle_lt_trans.

Lemma LtLeTrans: forall n1 n2 n3: t, Lt n1 n2 -> Le n2 n3 -> Lt n1 n3.
Proof Qclt_le_trans.

Lemma SplitEq: forall n1 n2: t,
    Le n1 n2 -> Le (neg n1) (neg n2) -> n1 = n2.
  unfold neg; intros n1 n2 Hle HleN.
  apply Qcle_antisym;
  [ | apply Qcopp_le_compat in HleN;
  rewrite (Qcopp_involutive n2), (Qcopp_involutive n1) in HleN ];
  assumption.
Qed.

Lemma MulAddDistrL: forall n1 n2 n3: t,
    mul (add n1 n2) n3 = add (mul n1 n3) (mul n2 n3).
  exact Qcmult_plus_distr_l.
Qed.

Lemma AddAssoc: forall n1 n2 n3: t,
    add n1 (add n2 n3) = add (add n1 n2) n3.
Proof Qcplus_assoc.

Lemma AddComm: forall n1 n2: t, add n1 n2 = add n2 n1.
Proof Qcplus_comm.

Lemma AddZL: forall n: t, add z n = n.
Proof Qcplus_0_l.

Lemma AddZR: forall n: t, add n z = n.
Proof Qcplus_0_r.

Lemma AddNeg: forall n: t, add n (neg n) = z.
Proof Qcplus_opp_r.

Lemma MulUL: forall n: t, mul u n = n.
Proof Qcmult_1_l.

Lemma MulZL: forall n: t, mul z n = z.
  intro; compute;
  apply Qc_is_canon; reflexivity.
Qed.

Lemma MulAssoc: forall n1 n2 n3: t,
    mul n1 (mul n2 n3) = mul (mul n1 n2) n3.
Proof Qcmult_assoc.

Lemma MulComm: forall n1 n2: t, mul n1 n2 = mul n2 n1.
Proof Qcmult_comm.

Lemma SubAdd: forall n1 n2: t, sub n1 n2 = add n1 (neg n2).
  intros; unfold sub, add, neg, Qcminus; trivial.
Qed.

Lemma AddEqEq: forall n1 n2 n3 n4: t,
    n1 = n2 -> n3 = n4 -> (add n1 n3) = (add n2 n4).
  intros n1 n2 n3 n4 h1 h2; rewrite h1, h2; trivial.
Qed.

Lemma AddLeLe: forall n1 n2 n3 n4: t,
    Le n1 n2 -> Le n3 n4 -> Le (add n1 n3) (add n2 n4).
Proof Qcplus_le_compat.

Lemma AddLtLt: forall n1 n2 n3 n4: t,
    Lt n1 n2 -> Lt n3 n4 -> Lt (add n1 n3) (add n2 n4).
Proof CoqAddOn.Qc.Qcplus_lt_compat.

Lemma AddEqLe: forall n1 n2 n3 n4: t,
    n1 = n2 -> Le n3 n4 -> Le (add n1 n3) (add n2 n4).
  intros n1 n2 n3 n4 h1 h2;
  assert (h3 := Qcle_refl n1); rewrite h1 in h3 at 2;
  exact (AddLeLe _ _ _ _ h3 h2).
Qed.

Lemma AddEqLt: forall n1 n2 n3 n4: t,
    n1 = n2 -> Lt n3 n4 -> Lt (add n1 n3) (add n2 n4).
Proof CoqAddOn.Qc.Qcplus_eqlt_compat.

Lemma AddLeLt: forall n1 n2 n3 n4: t,
    Le n1 n2 -> Lt n3 n4 -> Lt (add n1 n3) (add n2 n4).
  intros n1 n2 n3 n4 h; unfold Le in h;
  apply Qcle_lt_or_eq in h; destruct h as [h | h];
  [ exact (AddLtLt _ _ _ _ h) | exact (AddEqLt _ _ _ _ (Qc_is_canon _ _ h))].
Qed.

Lemma MulLe: forall n1 n2 n3: t,
    not (Le n1 z) -> Le n2 n3 -> Le (mul n1 n2) (mul n1 n3).
  intros n1 n2 n3 h1 h2.
  rewrite (MulComm n1 n2), (MulComm n1 n3).
  exact (Qcmult_le_compat_r n2 n3 n1 h2 (Qclt_le_weak _ _ (Qcnot_le_lt _ _ h1))).
Qed.

Lemma MulLt: forall n1 n2 n3: t,
    not (Le  n1 z) -> Lt n2 n3 -> Lt (mul n1 n2) (mul n1 n3).
  intros n1 n2 n3 h1 h2.
  rewrite (MulComm n1 n2), (MulComm n1 n3).
  exact (Qcmult_lt_compat_r n2 n3 n1 (Qcnot_le_lt _ _ h1) h2).
Qed.

Definition Ring := mk_rt z u add mul sub neg eq AddZL AddComm AddAssoc
  MulUL MulComm MulAssoc MulAddDistrL SubAdd AddNeg.

