let polCmp err p1Eqs p1Ineqs optP1 optP2 =
	match optP1, optP2 with
	| None, None -> None
	| None, Some p2 -> Some ("cmp:<\np1: bottom\np2" ^ (Pol.pr p2))
	| Some p1, None -> Some ("cmp:>\np1:\n" ^ (Pol.pr p1) ^ "\np2: bottom\n")
	| Some p1, Some p2 ->

	let chk pA pB c m =
		if Pol.check pA pB c then
			m
		else
			let aStr = "\npA:\n" ^ (Pol.pr pA)
			and bStr = "\npB:\n" ^ (Pol.pr pB)
			and cStr = "\ncert of pA in pB:\n" ^ (Cert.pr c) ^ "\n" in
			Some ("(cannot check certificate)" ^ aStr ^ bStr ^ cStr)
	in
	let diff =
		let iDiff = p1Ineqs - (List.length (IneqSet.list p2.Pol.ineqs))
		and eDiff = p1Eqs - (List.length (EqSet.list p2.Pol.eqs)) in
		if iDiff = 0 && eDiff = 0 then
			None
		else
			let eStr = string_of_int eDiff and iStr = string_of_int iDiff in
			Some ("eqs:" ^ eStr ^ " ineqs:" ^ iStr)
	in
	let cmp =
		let s = "\np1:\n" ^ (Pol.pr p1) ^ "\np2:\n" ^ (Pol.pr p2) in
		match Pol.isIncl p1 p2, Pol.isIncl p2 p1 with
		| Pol.NoIncl, Pol.NoIncl -> Some ("cmp:?" ^ s)
		| Pol.Incl w, Pol.NoIncl -> chk p1 p2 w (Some ("cmp:<" ^ s))
		| Pol.NoIncl, Pol.Incl w -> chk p2 p1 w (Some ("cmp:>" ^ s))
		| Pol.Incl w1, Pol.Incl w2 -> chk p1 p2 w1 (chk p2 p1 w2 None)
	in
	match err, diff, cmp with
	| None, None, None -> None
	| Some e, None, None
	| None, Some e, None
	| None, None, Some e -> Some e

	| Some e1, Some e2, None
	| Some e1, None, Some e2
	| None, Some e1, Some e2 -> Some (e1 ^ " " ^ e2)

	| Some e1, Some e2, Some e3 -> Some (e1 ^ " " ^ e2 ^ " " ^ e3)

let pbSz (init: int) (l: Pol.t option list): int =
	List.fold_left (
		fun a mp ->
			match mp with
			| None -> a
			| Some p -> a + (List.length p.Pol.eqs) + (List.length p.Pol.ineqs)
	) init l

let assign_out n tC pAEqs pAIneqs pA p l =
	let b = Oclock.gettime Oclock.thread_cputime in
	let p1 =
		match p with
		| None -> None
		| Some p -> Some (Pol.assignM p l)
	in
	let e = Oclock.gettime Oclock.thread_cputime in
	let err = polCmp None pAEqs pAIneqs pA p1 in
	Stat.assign n tC (Int64.sub e b) err (pbSz (List.length l) [p])

let chull_out n tC eqRes ineqRes mres mpl =
	let chull mpl = List.fold_left
		(fun a mp ->
			match mp, a with None, _ -> a
			| Some p, None -> Some p
			| Some p1, Some p2 ->
				let (p3,_,_) = Pol.join p2 p1 in
				Some p3)
		None mpl
	in
	let chullCert mpl = List.fold_left
		(fun a mp ->
			match mp, a with None, _ -> a
			| Some p, None -> Some p
			| Some p1, Some p2 ->
				let (p3, w2, w1) = Pol.join p2 p1 in
				if Pol.check p2 p3 w2 && Pol.check p1 p3 w1 then
					Some p3
				else
					failwith "chull_out")
		None mpl
	in
	let check mpl = List.fold_left
		(fun (a, e) mp ->
			match mp, a with None, _ -> (a, e)
			| Some p, None -> (Some p, e)
			| Some p1, Some p2 ->
				let (p3, c2, c1) = Pol.join p2 p1 in
				let r1 = Pol.check p1 p3 c1 in
				let r2 = Pol.check p2 p3 c2 in
				(Some p3, e + (if r1 then 0 else 1) + (if r2 then 0 else 1)))
		(None, 0) mpl
	in
	let b = Oclock.gettime Oclock.thread_cputime in
	let _ = chull mpl in
	let e = Oclock.gettime Oclock.thread_cputime in
	let _ = chullCert mpl in
	let eCert = Oclock.gettime Oclock.thread_cputime in
	let (p, err) = check mpl in
	let errE =
		if err = 0 then
			None
		else
			Some ("err:" ^ (string_of_int err))
	in
	let err = polCmp errE eqRes ineqRes mres p in
	Stat.chull n tC (Int64.sub e b) (Int64.sub eCert e) err (pbSz 0 mpl)

let incl_out n tC refans mp1 mp2 =
	let incl maybep1 maybep2 =
		match maybep1 with None -> (true, None)
		| Some p1 ->
			match maybep2 with None -> (false, None)
			| Some p2 ->
				match Pol.isIncl p1 p2 with
				| Pol.NoIncl -> (false, None)
				| Pol.Incl w -> (true, Some w)
	in
	let inclCert maybep1 maybep2 =
		match maybep1 with None -> true
		| Some p1 ->
			match maybep2 with None -> false
			| Some p2 ->
				match Pol.isIncl p1 p2 with
				| Pol.NoIncl -> false
				| Pol.Incl w -> Pol.check p1 p2 w
	in
	let check mp1 mp2 mw =
		match mp1, mp2, mw with
		| Some p1, Some p2, Some w -> Pol.check p1 p2 w
		| None, _, Some _ | _, None, Some _ -> false
		| _, _, None -> true
	in
	let b = Oclock.gettime Oclock.thread_cputime in
	let (r, mw) = incl mp1 mp2 in
	let e = Oclock.gettime Oclock.thread_cputime in
	let _ = inclCert mp1 mp2 in
	let eCert = Oclock.gettime Oclock.thread_cputime in
	let (aE, a) =
		if refans = r then (false, "") else (true, "ans:ko")
	in
	let (cE, c) =
		if check mp1 mp2 mw then (false, "") else (true, "cert:ko")
	in
	let err = if aE || cE then Some (a ^ " " ^ c) else None in
	Stat.incl n tC (Int64.sub e b) (Int64.sub eCert e) err (pbSz 0 [mp1; mp2])

let meet_out n tC p1 p1Eqs p1Ineqs p0 l =
	let rec meet p =
		function
		| [] -> Some p
		| c::t ->
			match Pol.add p c with
			| Some p1 -> meet p1 t
			| None -> None
	in
	let b = Oclock.gettime Oclock.thread_cputime in
	let p =
		match p0 with
		| None -> None
		| Some p -> meet p l
	in
	let e = Oclock.gettime Oclock.thread_cputime in
	let err = polCmp None p1Eqs p1Ineqs p1 p in
	Stat.meet n tC (Int64.sub e b) err (pbSz (List.length l) [p0])
