let assignC = ref Int64.zero and assignML = ref Int64.zero and assignE = ref 0
let chullC = ref Int64.zero and chullML = ref Int64.zero and chullE = ref 0
	and chullMLcert = ref Int64.zero
let inclC = ref Int64.zero and inclML = ref Int64.zero and inclE = ref 0
	and inclMLcert = ref Int64.zero
let meetC = ref Int64.zero and meetML = ref Int64.zero and meetE = ref 0

let chullSz: (int64 * int64) list ref = ref []
let assignSz: (int64 * int64) list ref = ref []
let inclSz: (int64 * int64) list ref = ref []
let meetSz: (int64 * int64) list ref = ref []

let rec szUpdate l sz c ml =
	match l with
	| [] -> if sz = 0 then [c, ml] else (Int64.zero, Int64.zero)::(szUpdate [] (sz-1) c ml)
	| (c1, ml1)::t ->
		if sz = 0 then
			(Int64.add c c1, Int64.add ml ml1)::t
		else
			(c1, ml1)::(szUpdate t (sz-1) c ml)

let consNb = ref []
let varNb = ref []

let rec nbUp i = function
	| [] -> if i = 0 then [1] else 0::(nbUp (i-1) [])
	| h::t -> if i = 0 then (h+1)::t else h::(nbUp (i-1) t)

let rec nbPr s i = function
	| [] -> s
	| h::t -> nbPr (s ^ (string_of_int i) ^ ": " ^ (string_of_int h) ^ "\n") (i+1) t

let rec nbAvg i a t = function
	| [] -> (float a) /. (float t)
	| h::l -> nbAvg (i+1) (a + i * h) (t+h) l

let consNbUp i = consNb := nbUp i !consNb
let varNbUp i = varNb := nbUp i !varNb

let consNbPr l = print_endline (nbPr "number of constraints:\n" 0 l)
let varNbPr l = print_endline (nbPr "number of variables:\n" 0 l)

let consNbAvg l =
	match l with
	| [] -> "average number of constraints: empty list"
	| _::t -> "average number of constraints: " ^ (string_of_float (nbAvg 1 0 0 t))

let varNbAvg l =
	match l with
	| [] -> "average number of variables: empty list"
	| _::t -> "average number of variables: " ^ (string_of_float (nbAvg 1 0 0 t))

let prUnit n = 
	let u = ["ns"; "us"; "ms"; "s"] in
	let prD d64 =
		let d = Int64.to_int d64 in
		if d >= 100 then
			"." ^ (string_of_int d)
		else if d >= 10 then
			".0" ^ (string_of_int d)
		else if d > 0 then
			".00" ^ (string_of_int d)
		else
			""
	in
	let rec pr i d = function
	| [] -> (Int64.to_string i) ^ (prD d) ^ "min"
	| h::t ->
		let ni = Int64.div i (Int64.of_int 1000) in
		if Int64.compare Int64.zero ni = 0 then
			(Int64.to_string i) ^ (prD d) ^ h
		else
			let nd = Int64.rem i (Int64.of_int 1000) in
			pr ni nd t
	in
	pr n Int64.zero u

let pr1 label c ml details =
	let sC = prUnit c and sML = prUnit ml in
	print_endline (label ^ "\tC:" ^ sC ^ "\tML:" ^ sML ^ " " ^ details)

let rec szPr a i l =
	match l with
	| [] -> a
	| (c, ml)::t ->
		let line = (string_of_int i) ^ " constraints:, Apron: " ^ (prUnit c) ^ ", libpoly: " ^ (prUnit ml) ^ "\n" in
		szPr (a ^ line) (i+1) t

let assign n tC tML errs sz =
	assignC := Int64.add !assignC tC;
	assignML := Int64.add !assignML tML;
	assignSz := szUpdate !assignSz sz tC tML;
	Printf.printf "(* Db.Libpoly ({ Db.prg = \"#\"; Db.op = Db.Assign; Db.n = %u }, Int64.of_string \"%Ld\"); *)\n" n tML;
	match errs with
	| None -> pr1 ("assign" ^ (string_of_int n)) tC tML ""
	| Some e ->
		assignE := !assignE + 1;
		pr1 ("assign" ^ (string_of_int n)) tC tML e

let chull n tC tML tMLcert errs sz =
	chullC := Int64.add !chullC tC;
	chullML := Int64.add !chullML tML;
	chullMLcert := Int64.add !chullMLcert tMLcert;
	chullSz := szUpdate !chullSz sz tC tML;
	Printf.printf "(* Db.Libpoly ({ Db.prg = \"#\"; Db.op = Db.Join; Db.n = %u }, Int64.of_string \"%Ld\"); *)\n" n tML;
	Printf.printf "(* Db.Check ({ Db.prg = \"#\"; Db.op = Db.Join; Db.n = %u }, Int64.of_string \"%Ld\"); *)\n" n tMLcert;
	match errs with
	| None -> pr1 ("chull" ^ (string_of_int n)) tC tML ""
	| Some e ->
		chullE := !chullE + 1;
		pr1 ("chull" ^ (string_of_int n)) tC tML e

let incl n tC tML tMLcert errs sz =
	inclC := Int64.add !inclC tC;
	inclML := Int64.add !inclML tML;
	inclMLcert := Int64.add !inclMLcert tMLcert;
	inclSz := szUpdate !inclSz sz tC tML;
	Printf.printf "(* Db.Libpoly ({ Db.prg = \"#\"; Db.op = Db.Incl; Db.n = %u }, Int64.of_string \"%Ld\"); *)\n" n tML;
	Printf.printf "(* Db.Check ({ Db.prg = \"#\"; Db.op = Db.Incl; Db.n = %u }, Int64.of_string \"%Ld\"); *)\n" n tMLcert;
	match errs with
	| None ->
		(* The following line is useless but removing it causes strange
		timing behaviour for the re benchmark.  This needs investigation. *)
		let e = " " ^ (Int64.to_string tML) ^ " " ^ (Int64.to_string tMLcert)
		in
		pr1 ("incl" ^ (string_of_int n)) tC tML e
	| Some e ->
		inclE := !inclE + 1;
		pr1 ("incl" ^ (string_of_int n)) tC tML e

let meet n tC tML errs sz =
	meetC := Int64.add !meetC tC;
	meetML := Int64.add !meetML tML;
	meetSz := szUpdate !meetSz sz tC tML;
	Printf.printf "(* Db.Libpoly ({ Db.prg = \"#\"; Db.op = Db.Meet; Db.n = %u }, Int64.of_string \"%Ld\"); *)\n" n tML;
	match errs with
	| None -> pr1 ("meet" ^ (string_of_int n)) tC tML ""
	| Some e ->
		meetE := !meetE + 1;
		pr1 ("meet" ^ (string_of_int n)) tC tML e

let summary () =
	let _pr label tC tML more =
		let s = label ^ ": Apron took " ^ (prUnit tC) ^
			", libpoly took " ^ (prUnit tML) ^ more
		in
		print_endline s
	in
	let pr label tC tML = _pr label tC tML "" in
	let prE label tC tML e =
		let more =
			if e = 0 then
				" (no errors)"
			else
				" (" ^ (string_of_int e) ^ " errors)"
		in
		_pr label tC tML more
	in
	let prCert label tC tML tCert e =
		let more =
			if e = 0 then
				" (no errors)"
			else
				" (" ^ (string_of_int e) ^ " errors)"
		in
		let chk =
			let chkIncr =
				let diff = Int64.sub tCert tML in
				let hundred = Int64.of_int 100 in
				Int64.div (Int64.mul hundred diff) tML
			in
			"(" ^ (prUnit tCert) ^ ", +" ^ (Int64.to_string chkIncr) ^ "%)"
		in
		_pr label tC tML (" " ^ chk ^ " " ^ more)
	in
	let totC =
		Int64.add !assignC (Int64.add !chullC (Int64.add !inclC !meetC))
	and totML =
		Int64.add !assignML (Int64.add !chullML (Int64.add !inclML !meetML))
	in
	prE "assign" !assignC !assignML !assignE;
	prCert "chull" !chullC !chullML !chullMLcert !chullE;
	prCert "incl" !inclC !inclML !inclMLcert !inclE;
	prE "meet" !meetC !meetML !meetE;
	pr "total" totC totML;
	consNbPr !consNb;
	varNbPr !varNb;
	print_endline (consNbAvg !consNb);
	print_endline (varNbAvg !varNb);
	print_endline (szPr "convex hull problem sizes\n" 0 !chullSz);
	print_endline (szPr "inclusion problem sizes\n" 0 !inclSz);
	print_endline (szPr "assignment problem sizes\n" 0 !assignSz);
	print_endline (szPr "meet problem sizes\n" 0 !meetSz)
